#include "GPIO.h"
#include "OWI.h"
#include "Software/OWI.h"
#include "Driver/Arduino.h"
#include "assert.h"
#include "benchmark.h"

// One-Wire Interface Remote Arduino (Master)

/*
 * Global Initialization
 */
// All possible addresses, ordered in ascending order by last two bytes
// To add more possible addresses, simply count up in binary
// Increase size of array accordingly
uint8_t Roms[16][7] = {
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x01},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x10},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x11},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x01, 0x00},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x01, 0x01},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x01, 0x10},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x01, 0x11},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x10, 0x00},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x10, 0x01},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x10, 0x10},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x10, 0x11},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x11, 0x00},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x11, 0x01},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x11, 0x10},
  {Arduino::FAMILY_CODE, 0xFF, 0x00, 0x00, 0x00, 0x11, 0x11}
};

// Initializing the array of scanned address IDs 
// increase size of array according to number of addresses in Roms array
uint8_t Ids[16] = {255};

// Initializing arduino master
Software::OWI<BOARD::D2> owi;
Arduino arduino(owi);

// Initializing helper variables
const byte numChars = 32;
char receivedChars[numChars]; // an array to store the received data
boolean newData = false;
boolean dataReady = false;
char cmd[32] = {0};
int state = 0;
int addr = 0;
int pin = 0;

/*
 * SETUP
 */
void setup()
{
  // Connect to Serial with Baud rate and print "Ready" on connect
  Serial.begin(9600);
  while (!Serial);

  Serial.println("Ready");
}

/*
 * LOOP
 */
void loop()
{
  // Scan serial interface periodically and parse incoming Messages
  recvWithEndMarker();
  parseData();

  // Decide, which command was sent and execute the called function
  if (dataReady) 
  {
    if (!strcmp(cmd,"temperature")) {
      owiGetTemperature(getRom(addr));      
    }
    else if (!strcmp(cmd,"humidity")) {
      owiGetHumidity(getRom(addr));
    }
    else if (!strcmp(cmd,"co2")) {
      owiGetCO2(getRom(addr));
    }
    else if (!strcmp(cmd,"tvoc")) {
      owiGetTVOC(getRom(addr));
    }
    else if (!strcmp(cmd, "pinMode")) {
      int mode;
      
      switch (state) {
        case 0:
          mode = INPUT;
          break;
        case 1:
          mode = OUTPUT;
          break;
        case 2:
          mode = INPUT_PULLUP;
          break;
      }
      
      if (addr == "Master") {
        pinMode(pin, mode);
      }
      else {
        owiPinMode(getRom(addr), pin, mode);
      }
    }
    else if (!strcmp(cmd, "digitalRead")) {
      if (addr == -1) {
        uint8_t val = digitalRead(pin);
        Serial.println(val);
      }
      else {
        uint8_t val = owiDigitalRead(getRom(addr), pin);
        Serial.println(val);
      }
    }
    else if (!strcmp(cmd, "digitalWrite")) {
      if (addr == -1) {
        digitalWrite(pin, state);
      }
      else {
        owiDigitalWrite(getRom(addr), pin, state);
      }
    }
    else if (!strcmp(cmd, "analogRead")) {
      if (addr == -1) {
        uint8_t val = analogRead(pin);
        Serial.println(val);
      }
      else {
        uint8_t val = owiAnalogRead(getRom(addr), pin);
        Serial.println(val);
      }
      
    }
    else if (!strcmp(cmd, "analogWrite")) {
      if (addr == -1) {
        analogWrite(pin, state);
      }
      else {
        owiAnalogWrite(getRom(addr), pin, state);
      }
    }
    else if (!strcmp(cmd, "numDigital")) {
      numDigitalPins(getRom(addr));
    }
    else if (!strcmp(cmd, "numAnalog")) {
      numAnalogPins(getRom(addr));
    }
    else if (!strcmp(cmd,"scan")) {
      scanRoms();
    }
  }
  cmd[32] = {0};
  state = 0;
  addr = 0;
  pin = 0;
  dataReady = false;
}


/**
 * Communication functions
 */

// Request temperature readout of called address
// Returns temperature on serial in °C
float owiGetTemperature(const uint8_t* ROM)
{
  arduino.rom(ROM);
  if (!arduino.convert_request(true)) return(-1000);

  delay(1000);

  arduino.read_scratchpad();
  float t = arduino.temperature();

  Serial.println(t);

  return(t);
}

// Request humidity readout of called address
// Teturns humidity on serial in percent
float owiGetHumidity(const uint8_t* ROM)
{
  arduino.rom(ROM);
  if (!arduino.convert_request(true)) return(-1000);

  delay(1000);

  arduino.read_scratchpad();
  float h = arduino.humidity();

  Serial.println(h);

  return(h);
}

// Request CO2 concentration readout of called address
// Returns CO2 concentration on serial in ppm
float owiGetCO2(const uint8_t* ROM)
{
  arduino.rom(ROM);
  if (!arduino.convert_request(true)) return(-1000);

  delay(1000);

  arduino.read_scratchpad();
  float co2 = arduino.co2();

  Serial.println(co2);

  return(co2);
}

// Request TVOC concentration readout of called address
// Returns TVOC concentraiton on serial in ppb
float owiGetTVOC(const uint8_t* ROM)
{
  arduino.rom(ROM);
  if (!arduino.convert_request(true)) return(-1000);

  delay(1000);

  arduino.read_scratchpad();
  float tvoc = arduino.tvoc();

  Serial.println(tvoc);

  return(tvoc);
}

// Change the pinMode of the given pin on a certain address
// Returns 0 with no error and -1 with error
uint8_t owiPinMode(const uint8_t* ROM, uint8_t pin, uint8_t mode)
{
  //INPUT = 0
  //OUTPUT = 1
  //INPUT_PULLUP = 2

  uint8_t val;
  arduino.rom(ROM);
  val = arduino.pinMode(pin, mode);
  Serial.println(val);
  return val;
}

// Request readout of a given digital pin on a certain address
bool owiDigitalRead(const uint8_t* ROM, uint8_t pin)
{
  bool val;
  arduino.rom(ROM);
  val = arduino.digitalRead(pin);
  Serial.println(val);
  return val;
}

// Write a 0 or a 1 to a given digital Pin on a certain address
// Returns 0 with no error and -1 with error
uint8_t owiDigitalWrite(const uint8_t* ROM, uint8_t pin, bool state)
{
  //LOW = 0
  //HIGH = 1
  uint8_t val;
  arduino.rom(ROM);
  val = arduino.digitalWrite(pin, state);
  Serial.println(val);
  return val;
}

// Request readout of given analog Pin on certain address
uint16_t owiAnalogRead(const uint8_t* ROM, uint8_t pin)
{
  uint16_t val;
  arduino.rom(ROM);
  val = arduino.analogRead(pin);
  Serial.println(val);
  return val;
}

// Write an analog value between 0-255 to a given pin on a certain address
// Returns 0 with no error and -1 with error
uint8_t owiAnalogWrite(const uint8_t* ROM, uint8_t pin, uint8_t duty)
{
  uint8_t val;
  arduino.rom(ROM);
  val = arduino.analogWrite(pin, duty);
  Serial.println(val);
  return val;
}

// Returns number of digital pins on certain address
uint8_t numDigitalPins(const uint8_t* ROM)
{
  uint8_t val;
  arduino.rom(ROM);
  val = arduino.num_digital_pins();
  Serial.println(val);
  return val;
}

// Returns number of analog pins on certain address
uint8_t numAnalogPins(const uint8_t* ROM)
{
  uint8_t val;
  arduino.rom(ROM);
  val = arduino.num_analog_inputs();
  Serial.println(val);
  return val;
}

// Scans for all addresses that were specified in the "Roms" array
// Returns all responding IDs to serial
void scanRoms()
{
  uint8_t rom[owi.ROM_MAX] = { 0 };
  uint8_t temp[owi.ROM_MAX] = { 0 };
  int8_t last = owi.FIRST;
  int id = 0;
  int cnt = 0;
  int i = 0;
  bool res = false;
  bool duplicate = false;

  for (int i=0; i<16; i++) {
    Ids[i] = 255;
  }

    do {
      id = 0;
      i = 0;
      last = owi.search_rom(Arduino::FAMILY_CODE, rom, last);

      // compare read rom with all roms in Roms array
      // if rom matches one address, try to write the corresponding ID in Ids array
      // if ID is already in ids, skip
      for (id; id<16; id++) {
        i = 0;
        do{
          temp[i] = Roms[id][i];
          if (temp[i] != rom[i]) {
            break;
          }
          i++;
        } while(i < 7);
                 
        if (i == 7) {
          if (Ids[id] != id) {
            Ids[id] = id;
            Serial.println(id);
          }
        }
      }
      cnt++;  
    } while (cnt<16);
}

// Returns the matching rom to a given ID
// Can be used as argument for all arduino functions
uint8_t* getRom(int id)
{
  static uint8_t temp[7];
  int i = 0;
  
  do{
    temp[i] = Roms[id][i];
    i++;
  } while(i < 7);
  
  return temp;
} 

// Check serial bus for new messages and save them in a chararray
// Sets "newData" flag to true
void recvWithEndMarker() {
 static byte ndx = 0;
 char endMarker = '\n';
 char rc;
 
 // if (Serial.available() > 0) {
           while (Serial.available() > 0 && newData == false) {
 rc = Serial.read();

 if (rc != endMarker) {
 receivedChars[ndx] = rc;
 ndx++;
 if (ndx >= numChars) {
 ndx = numChars - 1;
 }
 }
 else {
 receivedChars[ndx] = '\0'; // terminate the string
 ndx = 0;
 newData = true;
 }
 }
}

// Parses the received chararry
// Uses "," as separator and saves values in different variables
// Sets "dataReady" flag to 1 and resets "newData"
void parseData() {
  if (newData == true) {
    // split the data into its parts
    
  char * strtokIndx; // this is used by strtok() as an index
  
  strtokIndx = strtok(receivedChars,",");      // get the first part - the string
  strcpy(cmd, strtokIndx); // copy it to messageFromPC
  
  strtokIndx = strtok(NULL, ","); // this continues where the previous call left off
  
  if (String(strtokIndx) == "Master") {
    addr = -1;
  }else {
    addr = atoi(strtokIndx);     // convert this part to an integer
  }

  strtokIndx = strtok(NULL, ","); // this continues where the previous call left off
  pin = atoi(strtokIndx);     // convert this part to an integer
    
  strtokIndx = strtok(NULL, ","); // this continues where the previous call left off
  state = atoi(strtokIndx);  // convert this part to an integer
  
  newData = false;
  dataReady = true;
 }
}
